# FILE: mif.tcl
#
# Class for holding micromagnetic problem spec
#
# The input file should look like
#
#       # MIF 2.0
#       Specify Energy:UniformExchange: {
#           A 21e-12
#       }
#       Specify Energy:UniformAnisotropy: {
#           Type Uniaxial
#           K1   0.5e3
#           Axis 1 0 0
#       }
#
# This is evaluated inside a safe interpreter, so "unsafe" commands
# are not available
#
Oc_Class Oxs_Mif {
    private variable speckeys   ;# Ordered list of keys
    private array variable spec
    private variable misckeys   ;# Ordered list of keys
    private array variable misc

    Constructor { args } {
        eval $this Configure $args
        set speckeys {}
    }

    method SetSpec { key args } {
	set ac [llength $args]
	if {$ac != 1} {
	    return -code error \
		    "Bad argument count ($ac) in Specify block (KEY $key)"
	}
	if {[string first ":" $key]<0} {
	    # Append default instance name (empty string)
	    append key ":"
	}
	if {[info exists spec($key)]} {
	    return -code error "Key $key already in use"
	}
        lappend speckeys $key
        set spec($key) [lindex $args 0]
    }

    method UnsetSpec { key } {
	# If $key is {}, then clears out all specs.  This does not
        # raise an error even if there are no specs.
        #   If $key is not {}, then removes only that spec, from
        # both the spec array and the speckeys list.  This operation
        # will raise an error if the spec $key is not set.
        if {[string match {} $key]} {
            catch {unset spec}
            set speckeys {}
        } else {
            unset spec($key)
            set keyindex [lsearch -exact $speckeys $key]
            if {$keyindex>=0} {
                set speckeys [lreplace $speckeys $keyindex $keyindex]
            }
        }
    }
    # It might also be useful to have a 'ReplaceSpec' method
    # that does an Unset + Set without changing key order.

    method SetMisc { misclist } {
	if {[llength $misclist]%2 != 0} {
	    return -code error \
		    "Bad argument list in Miscellaneous block"
	}
	foreach {key value} $misclist {
	    lappend misckeys $key
	    set misc($key) $value
	}
    }

    method UnsetMisc { keys } {
	# If $keys is {}, then clears out misc.  This does not
        # raise an error even if misc does not exist.
        #   If $keys is not {}, then removes each item from
        # both the misc array and the misckeys list.  This operation
        # will raise an error if the misc $key is not set.
        if {[string match {} $keys]} {
            catch {unset misc}
            set misckeys {}
        } else {
	    foreach item $keys {
		unset misc($item)
		set itemindex [lsearch -exact $misckeys $item]
		if {$itemindex>=0} {
		    set misckeys [lreplace $misckeys $itemindex $itemindex]
		}
	    }
        }
    }

    method Clear {} {
	$this UnsetSpec {}
	$this UnsetMisc {}
    }

    method Read { safe_interp filename } {
        # Fill spec from filename.  A fresh safe interpreter should
	# be created each time this function is called.  This safe
	# interpreter must be a child of the current interpreter.

	# Empty out any previous specifications (should we do this?)
	$this Clear

	# Read file into a string.  Hopefully this isn't too big.
	set chan [open $filename]
	set filestr [read $chan]
	close $chan

	# Check that first line matches "# MIF 2.0"
	if {![regexp -- "^#\[ |\t\]*MIF\[ |\t\]*2.0\[ |\t\]*\n" \
		$filestr]} {
	    return -code error \
		    "Input file $filename not in MIF 2.0 format"
	}

	# Setup safe interpreter to source filestr
        interp alias $safe_interp Specify {} $this SetSpec
	interp alias $safe_interp ClearSpec {} $this UnsetSpec
        interp alias $safe_interp Miscellaneous {} $this SetMisc
	interp alias $safe_interp Report {} puts stderr

	# Source filestr
        set errcode [catch {$safe_interp eval $filestr} errmsg]
	global errorInfo errorCode
	foreach {ei ec} [list $errorInfo $errorCode] {break}

	# Remove main interp hooks
        interp alias $safe_interp Specify {}
	interp alias $safe_interp ClearSpec {}
        interp alias $safe_interp Miscellaneous {}

	if {$errcode} {
	    # Error occurred sourcing $filename
	    $this Clear      ;# Clear out partial problem spec
	    return -code error -errorinfo $ei -errorcode $ec \
		    "Error processing file $filename: $errmsg"
	}

    }

    method GetSpecKeys {} {
        return $speckeys
    }

    method GetSpecValue { name } {
        return $spec($name)
    }

    method GetMiscKeys {} {
	return $misckeys
    }

    method GetMiscValue { name } {
        return $misc($name)
    }

    method Dump {} {
        # Return data in MIF 2.0 file format.
        set specstr "# MIF 2.0\n"
        foreach k $speckeys {
            append specstr "[list Specify $k $spec($k)]\n"
        }
        return $specstr
    }

    Destructor {
    }
}
