/* FILE: eulerevolve.cc                 -*-Mode: c++-*-
 *
 * Concrete evolver class, using simple forward Euler steps
 *
 */

#include <float.h>

#include "nb.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "eulerevolve.h"
#include "key.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_EulerEvolve);

/* End includes */

// Constructor
Oxs_EulerEvolve::Oxs_EulerEvolve(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Evolver(name,newdtr,safe_interp,argstr),
    energy_accum_count_limit(25),
    next_timestep(0.),state_id(0),max_dm_dt(0.),dE_dt(0.),
    total_energy(0.),total_energy_change(0.),
    max_dm_dt_output(this,InstanceName(),"Max dm/dt","deg/ns",0,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    dE_dt_output(this,InstanceName(),"dE/dt","J/s",0,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    total_energy_output(this,InstanceName(),"Total energy","J",0,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    dm_dt_output(this,InstanceName(),"dm/dt","rad/s",1,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr),
    total_H_field_output(this,InstanceName(),"H","A/m",1,
		     &Oxs_EulerEvolve::UpdateDerivedOutputs,newdtr)
{
  // Process arguments
  allowed_error_rate = GetRealInitValue("error_rate",-1);
  if(allowed_error_rate>0.0) {
    allowed_error_rate *= PI*1e9/180.; // Convert from deg/ns to rad/s
  }
  allowed_absolute_step_error
    = GetRealInitValue("absolute_step_error",0.2);
  if(allowed_absolute_step_error>0.0) {
    allowed_absolute_step_error *= PI/180.; // Convert from deg to rad
  }
  allowed_relative_step_error
    = GetRealInitValue("relative_step_error",0.2);

  step_headroom = GetRealInitValue("step_headroom",0.85);
  if(step_headroom<=0.) {
    string msg="Invalid initialization detected for object "
      + string(InstanceName())
      + ": step_headroom value must be bigger than 0.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  gamma = GetRealInitValue("gamma",2.21e5);

  do_precess = GetIntInitValue("do_precess",1);

  start_dm = GetRealInitValue("start_dm",0.01);

  alpha = GetRealInitValue("alpha",0.5);

  VerifyAllInitArgsUsed();
}

BOOL Oxs_EulerEvolve::Init()
{
  next_timestep=0.0;
  state_id=0;   // Mark as invalid state
  max_dm_dt=-1; // Dummy value
  dE_dt=0;      // Dummy value
  total_energy=0.0;
  total_energy_change=0.0; // Dummy value
  energy_accum_count=energy_accum_count_limit; // Force cold count
  /// on first pass
  return 1;
}


Oxs_EulerEvolve::~Oxs_EulerEvolve()
{}


void Oxs_EulerEvolve::Calculate_dm_dt
(const Oxs_Mesh& mesh_,
 const Oxs_MeshValue<REAL8m>& Ms_,
 const Oxs_MeshValue<ThreeVector>& mxH_,
 const Oxs_MeshValue<ThreeVector>& spin_,
 Oxs_MeshValue<ThreeVector>& dm_dt_,
 REAL8m& max_dm_dt_,REAL8m& dE_dt_,REAL8m& min_timestep_) const
{ // Imports: mesh_, Ms_, mxH_, spin_
  // Exports: dm_dt_, max_dm_dt_, dE_dt_
  UINT4m size = mesh_.Size();  // Assume all imports are compatible
  REAL8m coef1 = -fabs(gamma);
  REAL8m coef2 = -fabs(alpha);
  ThreeVector scratch;
  REAL8m max_dm_dt_sq=0.0;
  UINT4m max_index=0;
  REAL8m dm_dt_sq;
  REAL8m dE_dt_sum=0.0;
  dm_dt_.AdjustSize(&mesh_);
  for(UINT4m i=0;i<size;i++) {
    if(Ms_[i]==0) {
      dm_dt_[i].Set(0.0,0.0,0.0);
    } else {
      scratch = mxH_[i];
      dE_dt_sum += scratch.MagSq() * Ms_[i] * mesh_.Volume(i);
      scratch *= coef1;
      if(do_precess) {
	dm_dt_[i] = scratch;
      } else {
	dm_dt_[i].Set(0.0,0.0,0.0);
      }
      scratch ^= spin_[i]; // -|gamma|((mxH)xm)
      scratch *= coef2;  // |alpha.gamma|((mxH)xm) = -|alpha.gamma|(mx(mxH))
      dm_dt_[i] += scratch;
      dm_dt_sq = dm_dt_[i].MagSq();
      if(dm_dt_sq>max_dm_dt_sq) {
	max_dm_dt_sq=dm_dt_sq;
	max_index = i;
      }
    }
  }
  max_dm_dt_ = sqrt(max_dm_dt_sq);
  dE_dt_ = -1 * MU0 * fabs(gamma*alpha) * dE_dt_sum;

  // Get bound on smallest stepsize that would actually
  // change spin new_max_dm_dt_index:
  REAL8m min_ratio = DBL_MAX/2.;
  if(fabs(dm_dt_[max_index].x)>=1.0 ||
     min_ratio*fabs(dm_dt_[max_index].x) > fabs(spin_[max_index].x)) {
    min_ratio = fabs(spin_[max_index].x/dm_dt_[max_index].x);
  }
  if(fabs(dm_dt_[max_index].y)>=1.0 ||
     min_ratio*fabs(dm_dt_[max_index].y) > fabs(spin_[max_index].y)) {
    min_ratio = fabs(spin_[max_index].y/dm_dt_[max_index].y);
  }
  if(fabs(dm_dt_[max_index].z)>=1.0 ||
     min_ratio*fabs(dm_dt_[max_index].z) > fabs(spin_[max_index].z)) {
    min_ratio = fabs(spin_[max_index].z/dm_dt_[max_index].z);
  }
  min_timestep_ = min_ratio * REAL8_EPSILON;

  return;
}


BOOL
Oxs_EulerEvolve::Step(const Oxs_Driver& driver,
		      Oxs_ConstKey<Oxs_SimState> current_state,
		      Oxs_Key<Oxs_SimState>& next_state)
{
  const REAL8m max_step_increase = 1.25;
  const REAL8m max_step_decrease = 0.5;

  UINT4m size,i; // Mesh size and indexing variable

  const Oxs_SimState& cstate = current_state.GetReadReference();
  Oxs_SimState& workstate = next_state.GetWriteReference();

  if(cstate.mesh->Id() != workstate.mesh->Id()) {
    throw Oxs_Ext::Error
      ("Oxs_EulerEvolve::Step: Oxs_Mesh not fixed across steps.");
  }

  REAL8m stepsize = next_timestep;
  if( cstate.Id() != state_id || stepsize<=0.0) {
    // Cached data out of data
    director->GetEnergyDensity(cstate,energy,&mxH,NULL);
    Calculate_dm_dt(*(cstate.mesh),*(cstate.Ms),
		    mxH,cstate.spin,
		    dm_dt,max_dm_dt,dE_dt,timestep_lower_bound);
    if(start_dm < sqrt(DBL_MAX/4) * max_dm_dt) {
      stepsize = start_dm / max_dm_dt;
    } else {
      stepsize = sqrt(DBL_MAX/4);
    }
    state_id = cstate.Id();
  }

  // Negotiate with driver over size of next step
  workstate.last_timestep=stepsize;
  if(stepsize<timestep_lower_bound)
    workstate.last_timestep=timestep_lower_bound;
  driver.FillStatePartB(workstate);
  BOOL forcestep=0;
  if(workstate.last_timestep>stepsize) {
    // Either driver wants to force this stepsize,
    // or else suggested stepsize is smaller than
    // timestep_lower_bound.
    forcestep=1;
  }
  stepsize = workstate.last_timestep;

  workstate.elapsed_time = cstate.elapsed_time + stepsize;
  workstate.iteration_count = cstate.iteration_count + 1;
  workstate.stage_iteration_count = cstate.stage_iteration_count + 1;

  // Put new spin configuration in next_state
  workstate.spin.AdjustSize(workstate.mesh); // Safety
  size = workstate.spin.Size();
  ThreeVector tempspin;
  for(i=0;i<size;++i) {
    tempspin = dm_dt[i];
    tempspin *= stepsize;

    // For improved accuracy, adjust step vector so that
    // to first order m0 + adjusted_step = v/|v| where
    // v = m0 + step.
    REAL8m adj = 0.5 * tempspin.MagSq();
    tempspin -= adj*cstate.spin[i];
    tempspin *= 1.0/(1.0+adj);

    tempspin += cstate.spin[i];
    tempspin.MakeUnit();
    workstate.spin[i] = tempspin;
  }
  const Oxs_SimState& nstate
    = next_state.GetReadReference();  // Release write lock

  //  Calculate delta E
  Oxs_MeshValue<ThreeVector>* H_ptr = NULL;
  if(total_H_field_output.GetCacheRequestCount()>0) {
    H_ptr = &total_H_field_output.cache.value;
    total_H_field_output.cache.state_id = 0;
  }
  director->GetEnergyDensity(nstate,new_energy,&mxH,H_ptr);
  if(H_ptr!=NULL) {
    total_H_field_output.cache.state_id = nstate.Id();
  }

  REAL8m dE=0.0;
  REAL8m var_dE=0.0;
  for(i=0;i<size;++i) {
    REAL8m vol = nstate.mesh->Volume(i);
    REAL8m e = energy[i];
    REAL8m new_e = new_energy[i];
    dE += (new_e - e) * vol;
    var_dE += (new_e*new_e + e*e)*vol*vol;
  }
  var_dE *= 256*REAL8_EPSILON*REAL8_EPSILON/3.; // Variance, assuming
  /// error in each energy[i] term is independent, uniformly
  /// distributed, 0-mean, with range +/- 16*REAL8_EPSILON*energy[i].
  /// It would probably be better to get an error estimate directly
  /// from each energy term.

  /**/
  workstate.delta_E = dE; // HACK ALERT!!! workstate is a pointer
  /// into the locked-down nstate, which is not suppose to be
  /// changed.  Hopefully this problem will go away when simstate
  /// evolves into a real container, that anybody can stick stuff
  /// into (but not pull out or modify the contents once inserted).
  /**/

  // Get error estimate.  See step size adjustment discussion in
  // MJD Notes II, p72 (18-Jan-2001).
  REAL8m new_max_dm_dt,new_dE_dt,new_timestep_lower_bound;
  Calculate_dm_dt(*(nstate.mesh),*(nstate.Ms),
		  mxH,nstate.spin,new_dm_dt,
		  new_max_dm_dt,new_dE_dt,new_timestep_lower_bound);
  REAL8m max_error=0;
  for(i=0;i<size;++i) {
    ThreeVector temp = dm_dt[i];
    temp -= new_dm_dt[i];
    REAL8m temp_error = temp.MagSq();
    if(temp_error>max_error) max_error = temp_error;
  }
  max_error = sqrt(max_error)/2.0; // Actual (local) error
  /// estimate is max_error * stepsize

  // Check step and adjust next_timestep.  The relative error
  // check is a bit fudged, because rather than limiting the
  // relative error uniformly across the sample, we limit it
  // only at the position that has the maximum absolute error
  // (i.e., max_error is max *absolute* error).  I haven't
  // tested to see if uniformly limiting relative error is
  // workable (it might be too restrictive for most purposes),
  // but the present setup seems to solve the problem of convergence
  // near equilibrium.  -mjd, 2001-02-23.
  // NOTE: Since all three error controls (error_rate,
  //  absolute_step_error, and relative_step_error) assume error
  //  grows linearly with step size, we can check up front to see
  //  which control is most restrictive, store that constraint in
  //  working_allowed_error, and then adjust the step size without
  //  regard to which control is being exercised.
  REAL8m working_allowed_error
    = max_step_increase*max_error/step_headroom;
  if(allowed_error_rate>=0.
     && working_allowed_error>allowed_error_rate) {
    working_allowed_error=allowed_error_rate;
  }
  if(allowed_absolute_step_error>=0.
     && stepsize*working_allowed_error>allowed_absolute_step_error) {
    working_allowed_error=allowed_absolute_step_error/stepsize;
  }
  if(allowed_relative_step_error>=0.
     && working_allowed_error>allowed_relative_step_error*max_dm_dt) {
    working_allowed_error = allowed_relative_step_error * max_dm_dt;
  }
  if(!forcestep) {
    next_timestep=1.0;  // Size relative to current step
    if(max_error>working_allowed_error) {
      next_timestep = step_headroom*working_allowed_error/max_error;
    } else if(dE>REAL8_EPSILON*fabs(total_energy)
	      && dE>0 && dE*dE>4*var_dE) {
      // Energy check
      next_timestep=0.5;
    }
    if(next_timestep<1.0) {
      // Reject step
      if(next_timestep<max_step_decrease)
	next_timestep=max_step_decrease;
      next_timestep *= stepsize;
      return 0;
    }
  }

  // Otherwise, accept step.  Calculate next step using
  // estimate of step size that would just meet the error
  // restriction (with "headroom" safety margin).
  state_id=0; // Safety
  next_timestep = max_step_increase;
  if(next_timestep*max_error>step_headroom*working_allowed_error) {
    next_timestep = step_headroom*working_allowed_error/max_error;
  }
  if(next_timestep<max_step_decrease)
    next_timestep=max_step_decrease;
  next_timestep *= stepsize;
  dm_dt.Swap(new_dm_dt);
  max_dm_dt=new_max_dm_dt;
  timestep_lower_bound=new_timestep_lower_bound;
  dE_dt=new_dE_dt;
  if(energy_accum_count<energy_accum_count_limit) {
    total_energy += dE;
    ++energy_accum_count;
  } else {
    total_energy=0.0;
    for(i=0;i<size;++i) {
      total_energy += new_energy[i] * nstate.mesh->Volume(i);
    }
    energy_accum_count=0;
  }
  total_energy_change = dE;
  energy.Swap(new_energy);
  state_id = nstate.Id();

  return 1;  // Good step
}

void Oxs_EulerEvolve::UpdateDerivedOutputs(const Oxs_SimState& state)
{ // This routine fills all the Oxs_EulerEvolve Oxs_ScalarOutput's to
  // the appropriate value based on the import "state".  Also any of
  // Oxs_VectorOutput's that have CacheRequest enabled are filled. 
  max_dm_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = total_energy_output.cache.state_id = 0;
  if(state.Id()==state_id) {
    // Use previously calculated values
    max_dm_dt_output.cache.value=max_dm_dt*(180e-9/PI);
             /// Convert from radians/second to deg/ns
    dE_dt_output.cache.value=dE_dt;
    total_energy_output.cache.value=total_energy;
    if(dm_dt_output.GetCacheRequestCount()>0) {
      dm_dt_output.cache.value=dm_dt;
      dm_dt_output.cache.state_id=state_id;
    }
  } else {
    // Calculate from scratch
    UINT4m size=state.mesh->Size();
    Oxs_MeshValue<ThreeVector>* H_ptr = NULL;
    if(total_H_field_output.GetCacheRequestCount()>0
       && total_H_field_output.cache.state_id != state.Id()) {
      H_ptr = &total_H_field_output.cache.value;
      total_H_field_output.cache.state_id = 0;
    }
    director->GetEnergyDensity(state,new_energy,&mxH,H_ptr);
    if(total_H_field_output.GetCacheRequestCount()>0
       && total_H_field_output.cache.state_id != state.Id()) {
      total_H_field_output.cache.state_id = state.Id();
    }
    REAL8m energy_sum=0.;
    for(UINT4m i=0;i<size;++i) {
      energy_sum += new_energy[i] * state.mesh->Volume(i);
    }
    total_energy_output.cache.value=energy_sum;

    Oxs_MeshValue<ThreeVector>* dm_dt_ptr = &dm_dt;
    if(dm_dt_output.GetCacheRequestCount()>0
       && dm_dt_output.cache.state_id != state.Id()) {
      dm_dt_ptr = &dm_dt_output.cache.value;
      dm_dt_output.cache.state_id=0;
    }
    REAL8m dummy_value;
    Calculate_dm_dt(*(state.mesh),*(state.Ms),mxH,state.spin,
		    *dm_dt_ptr,
		    max_dm_dt_output.cache.value,
		    dE_dt_output.cache.value,dummy_value);
    if(dm_dt_output.GetCacheRequestCount()>0
       && dm_dt_output.cache.state_id != state.Id()) {
      dm_dt_output.cache.state_id=state.Id();
    }
    max_dm_dt_output.cache.value *= (180e-9/PI);
      /// Convert from radians/second to deg/ns
  }
  max_dm_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = total_energy_output.cache.state_id = state.Id();

  if(total_H_field_output.GetCacheRequestCount()>0 
     && total_H_field_output.cache.state_id != state.Id()) {
    director->GetEnergyDensity(state,new_energy,NULL,
			       &total_H_field_output.cache.value);
    total_H_field_output.cache.state_id = state.Id();
  }
}
