/* FILE: uniformexchange.cc            -*-Mode: c++-*-
 *
 * Uniform 6 neighbor exchange energy on rectangular mesh,
 * derived from Oxs_Energy class.
 *
 */

#include "nb.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"
#include "uniformexchange.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UniformExchange);

/* End includes */


// Constructor
Oxs_UniformExchange::Oxs_UniformExchange(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("A",1);
  A=Nb_Atof((*FindInitValue("A"))[0].c_str());
  DeleteInitValue("A");

  VerifyAllInitArgsUsed();
}

Oxs_UniformExchange::~Oxs_UniformExchange()
{}

void Oxs_UniformExchange::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms_inverse = *(state.Ms_inverse);

  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(state.mesh);
  if(mesh==NULL) {
    string msg="Import mesh to Oxs_UniformExchange::GetEnergyAndField()"
      " routine of object " +  string(InstanceName())
      + " is not an Oxs_RectangularMesh object.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  UINT4m xdim = mesh->DimX();
  UINT4m ydim = mesh->DimY();
  UINT4m zdim = mesh->DimZ();
  UINT4m xydim = xdim*ydim;

  REAL8m wgtx = -A/(mesh->EdgeLengthX()*mesh->EdgeLengthX());
  REAL8m wgty = -A/(mesh->EdgeLengthY()*mesh->EdgeLengthY());
  REAL8m wgtz = -A/(mesh->EdgeLengthZ()*mesh->EdgeLengthZ());

  const REAL8m hcoef = -2/MU0;

  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address

	REAL8m Msii = Ms_inverse[i];
	if(Msii == 0.0) {
	  energy[i]=0.0;
	  field[i].Set(0.,0.,0.);
	  continue;
	}

	ThreeVector base = spin[i];

	ThreeVector sum(0.,0.,0.);
	if(x>0) {
	  if(Ms_inverse[i-1]!=0.0) sum = (spin[i-1] - base);
	}
	if(x<xdim-1) {
	  if(Ms_inverse[i+1]!=0.0) sum += (spin[i+1] - base);
	}
	sum *= wgtx;

	ThreeVector temp(0.,0.,0.);
	if(y>0) {
	  UINT4m j = i-xdim;
	  if(Ms_inverse[j]!=0.0) temp = (spin[j] - base);
	}
	if(y<ydim-1) {
	  UINT4m j = i+xdim;
	  if(Ms_inverse[j]!=0.0) temp += (spin[j] - base);
	}
	sum += wgty*temp;

	temp.Set(0.,0.,0.);
	if(z>0) {
	  UINT4m j = i-xydim;
	  if(Ms_inverse[j]!=0.0) temp = (spin[j] - base);
	}
	if(z<zdim-1) {
	  UINT4m j = i+xydim;
	  if(Ms_inverse[j]!=0.0) temp += (spin[j]- base);
	}
        sum += wgtz*temp;

	field[i] = (hcoef*Msii) * sum;
	energy[i] = (sum * base);
      }
    }
  }
}









